local DEBUG = nil

--------------------------------------------------------------------------------
-- Locale
--

local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

--------------------------------------------------------------------------------
-- DKP-type specific variables
--

local FLASK_MULTIPLIER = 0.4
local ELIXIR_MULTIPLIER = 0.2
local FOOD_MULTIPLIER = 0.1

local flasks = {
	-- WotLK Flasks
	["Flask of Endless Rage"] = true,
	["Flask of Pure Mojo"] = true,
	["Flask of Stoneblood"] = true,
	["Flask of the Frost Wyrm"] = true,
	["Lesser Flask of Toughness"] = true,

	-- TBC Flasks
	["Flask of Arcane Fortification"] = true,
	["Chromatic Resistance"] = true,
	["Distilled Wisdom"] = true,
	["Flask of Fortification"] = true,
	["Flask of Mighty Restoration"] = true,
	["Flask of Relentless Assault"] = true,
	["Flask of Shadow Fortification"] = true,
	["Supreme Power"] = true,
	["Flask of Pure Death"] = true,
	["Flask of the Titans"] = true,
	["Supreme Power of Shattrath"] = true,
	["Fortification of Shattrath"] = true,
	["Mighty Restoration of Shattrath"] = true,
	["Relentless Assault of Shattrath"] = true,
	["Pure Death of Shattrath"] = true,
	["Blinding Light of Shattrath"] = true,
	["Unstable Flask of the Bandit"] = true,
	["Unstable Flask of the Beast"] = true,
	["Unstable Flask of the Elder"] = true,
	["Unstable Flask of the Sorcerer"] = true,
	["Unstable Flask of the Soldier"] = true,
	["Unstable Flask of the Physician"] = true,
}

local elixirs = {
	-- WotLK Elixirs
	["Accuracy"] = true,
	["Armor Piercing"] = true,
	["Deadly Strikes"] = true,
	["Expertise"] = true,
	["Lightning Speed"] = true,
	["Mighty Agility"] = true,
	["Mighty Strength"] = true,
	["Protection"] = true,
	["Guru's Elixir"] = true,
	["Spellpower Elixir"] = true,
	["Wrath Elixir"] = true,

	["Mighty Defense"] = true,
	["Elixir of Mighty Fortitude"] = true,
	["Mighty Mana Regeneration"] = true,
	["Mighty Thoughts"] = true,
	["Elixir of Spirit"] = true,

	-- TBC Elixirs
	["Healing Power"] = true,
	["Elixir of Draenic Wisdom"] = true,
	["Elixir of Brute Force"] = true,
	["Fel Strength Elixir"] = true,
	["Empowerment"] = true,
	["Greater Mana Regeneration"] = true,
	["Major Shadow Power"] = true,
	["Major Armor"] = true,
	["Elixir of Ironskin"] = true,
	["Major Agility"] = true,
	["Earthen Elixir"] = true,
	["Major Frost Power"] = true,
	["Elixir of Mastery"] = true,
	["Elixir of Major Fortitude"] = true,
	["Major Strength"] = true,
	["Adept's Elixir"] = true,
	["Onslaught Elixir"] = true,
	["Greater Arcane Elixir"] = true,
	["Elixir of the Mongoose"] = true,
	["Greater Firepower"] = true,
	["Shadow Power"] = true,
	["Elixir of the Giants"] = true,
	["Arcane Elixir"] = true,
	["Elixir of Demonslaying"] = true,
}

-- Basically a copy of the list I made for NRT.
local ignoredItems = {
	[30316] = true, -- Devastation
	[30317] = true, -- Cosmic Infuser
	[30312] = true, -- Infinity Blade
	[30318] = true, -- Netherstrand Longbow
	[30313] = true, -- Staff of Disintegration
	[30311] = true, -- Warp Slicer
	[30320] = true, -- Bundle of Nether Spikes
	[30314] = true, -- Phaseshift Bulwark
	[30319] = true, -- Nether Spikes
	[29434] = true, -- Badge of Justice
	[22450] = true, -- Void Crystal
	[20725] = true, -- Nexus Crystal

	[32897] = true, -- Mark of the Illidari
	[32428] = true, -- Heart of Darkness
	[34664] = true, -- Sunmote

	[32230] = true, -- Shadowsong Amethyst
	[32249] = true, -- Seaspray Emerald
	[32231] = true, -- Pyrestone
	[32229] = true, -- Lionseye
	[32228] = true, -- Empyrean Sapphire
	[32227] = true, -- Crimson Spinel
	
	[36919] = true, -- Cardinal Ruby
	[36925] = true, -- Majestic Zircon
	[36922] = true, -- King's Amber
	[36928] = true, -- Dreadstone
	[36934] = true, -- Eye of Zul
	[36931] = true, -- Ametrine

	[40752] = true, -- Emblem of Heroism
	[40753] = true, -- Emblem of Valor
	[45624] = true, -- Emblem of Conquest
	[47241] = true, -- Emblem of Triumph
	[34057] = true, -- Abyss Crystal
}

local icon = LibStub("LibDBIcon-1.0")

local _G = getfenv(0)
local db = nil

local halfHour = 60 * 30
local oneHour = halfHour * 2

local handle_HourTicker = nil
local handle_MinuteTicker = nil
local handle_Broadcast = nil
local handle_BidTimer = nil

local cache_AttendancePercent = {}

local guildRanks = {}
local guildMemberRanks = {}
local guildMemberClasses = {}
local guildClasses = {}
local guildMemberList = {}

local lootDialogOpen = nil
local lootQueue = {}
local updateDisplayedNames = true
local takingBids = nil
local tmpBidTable = {}
local bidders = {}
local playerName = UnitName("player")
local chatHandlers = {}

local hexColors = {}
local wtfColor = "|cffa0a0a0%s|r"
for k, v in pairs(RAID_CLASS_COLORS) do
	hexColors[k] = "|cff" .. string.format("%02x%02x%02x", v.r * 255, v.g * 255, v.b * 255) .. "%s|r"
end
local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = guildMemberList[key] or select(2, UnitClass(key)) or "WTF"
		if not class or class == "WTF" then
			return wtfColor:format(key)
		else
			self[key] = hexColors[class]:format(key)
			return self[key]
		end
	end
})

--------------------------------------------------------------------------------
-- Utility functions
--

local function print(t)
	DEFAULT_CHAT_FRAME:AddMessage("BDKP: " .. t)
end

local function debug(t)
	if DEBUG then print("[DEBUG] " .. t) end
end

local function broadcastMulti()
	debug("broadcastMulti()")
	local dkp = BananaDKP:Serialize(db.dkp)
	local att = {}
	for k, v in pairs(db.attendance) do
		if not v:find("1") then
			att[k] = "-"
		elseif not v:find("0") then
			att[k] = "!"
		else
			att[k] = v
		end
	end
	local attendance = BananaDKP:Serialize(att)
	-- Broadcast attendance first, since it takes aaaages to send.
	BananaDKP:SendCommMessage("BDKP-ATT-M", attendance, "GUILD")
	BananaDKP:SendCommMessage("BDKP-DKP-M", dkp, "GUILD")
	wipe(att)
end

local function broadcastSingle(name)
	debug("broadcastSingle(" .. name .. ")")
	if db.attendance[name] then
		local attendance = BananaDKP:Serialize({ name, db.attendance[name] })
		BananaDKP:SendCommMessage("BDKP-ATT-S", attendance, "GUILD")
	end
	if db.dkp[name] then
		local dkp = BananaDKP:Serialize({ name, db.dkp[name] })
		BananaDKP:SendCommMessage("BDKP-DKP-S", dkp, "GUILD")
	end
end

-- Thanks, Wobin.
local function round(n, shift)
	shift = 10 ^ shift
	return math.floor((n * shift) + 0.5) / shift
end

local function getDKPMultiplier(player)
	local multiplier = 1
	local buffIndex = 1
	local buff = UnitBuff(player, buffIndex)
	while buff do
		if flasks[buff] then
			multiplier = multiplier + FLASK_MULTIPLIER
		elseif buff == L["Well Fed"] or buff == L["\"Well Fed\""] then
			multiplier = multiplier + FOOD_MULTIPLIER
		elseif elixirs[buff] then
			multiplier = multiplier + ELIXIR_MULTIPLIER
		end
		buffIndex = buffIndex + 1
		buff = UnitBuff(player, buffIndex)
	end
	return multiplier
end

local function getAttendancePercent(name)
	if cache_AttendancePercent[name] then return cache_AttendancePercent[name] end
	if not db.attendance[name] then return 0 end
	local total = 0
	for k in db.attendance[name]:gmatch("%d") do
		if k == "1" then total = total + 1 end
	end
	local p = (total / db.attendanceAverageOver) * 100
	cache_AttendancePercent[name] = p
	return p
end

local function getDKP(name)
	if not name then return end
	if not db.dkp[name] then return end
	return round(db.dkp[name], 2)
end

local function getDKPPercent(name)
	local p = getAttendancePercent(name)
	if p >= 60 then return 100
	else return (1.5 * p) + 10 end
end

local function getModifiedDKP(name)
	if not name then return end
	if not db.dkp[name] then return end
	local p = getDKPPercent(name)
	if p == 100 then return round(db.dkp[name], 2)
	else return round((db.dkp[name] * p) / 100, 2)
	end
end

local function dkpSorter(a, b)
	local aDKP = getModifiedDKP(a) or 0
	local bDKP = getModifiedDKP(b) or 0
	return aDKP > bDKP and true or false
end

local function itemLooted(player, itemLink)
	if db.mode == 2 then return end
	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink})
	else
		lootDialogOpen = true
		StaticPopupDialogs["BananaDKPItemQuery"].text = L["%s looted %s, click Add to add it to his loot table."]:format(coloredName[player], itemLink)
		local popup = StaticPopup_Show("BananaDKPItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end
local function timerLoot(tbl) itemLooted(unpack(tbl)) end
local function popLootQueue()
	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		BananaDKP:ScheduleTimer(timerLoot, 1, n)
	end
end
local function ignoreItem()
	lootDialogOpen = nil
	popLootQueue()
end

local function chatFilter(self, event, ...) return chatHandlers[event](...) end

--------------------------------------------------------------------------------
-- Options
--

BananaDKP = LibStub("AceAddon-3.0"):NewAddon("BananaDKP", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0")
local BananaDKP = BananaDKP

local function notHost() return db.mode ~= 1 end
local options = {
	type = "group",
	handler = BananaDKP,
	args = {
		start = {
			type = "execute",
			name = L["Start"],
			desc = L["Start the automatic hour counter."],
			func = "StartCounter",
			order = 101,
			hidden = notHost,
			disabled = function() return not UnitInRaid("player") or handle_HourTicker end,
		},
		stop = {
			type = "execute",
			name = L["Stop"],
			desc = L["Stop the automatic hour counter."],
			func = "StopCounter",
			disabled = function() return not handle_HourTicker end,
			hidden = notHost,
			order = 103,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
			hidden = notHost,
		},
		manualHeader = {
			type = "header",
			name = L["Manual"],
			order = 200,
			hidden = notHost,
		},
		plusone = {
			type = "text",
			name = L["Increment"],
			desc = L["Increment the hour counter for a player or the entire raid+sitout list (*) by one."],
			usage = L["<player or *>"],
			get = false,
			set = "IncrementHourCounter",
			order = 201,
			hidden = notHost,
		},
		minusone = {
			type = "text",
			name = L["Decrement"],
			desc = L["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."],
			usage = L["<player or *>"],
			get = false,
			set = "DecrementHourCounter",
			order = 202,
			hidden = notHost,
		},
		spacer2 = {
			type = "header",
			name = " ",
			order = 250,
			hidden = notHost,
		},
		optionsHeader = {
			type = "header",
			name = L["Options"],
			order = 300,
			hidden = notHost,
		},
		role = {
			type = "group",
			name = L["Role"],
			desc = L["Sets your role, either client or server."],
			args = {
				client = {
					type = "toggle",
					name = L["Client"],
					desc = L["Sets this BananaDKP to act as a client."],
					get = function() return db.mode == 2 end,
					set = function(v)
						db.mode = v and 2 or 1
					end,
				},
				server = {
					type = "toggle",
					name = L["Server"],
					desc = L["Sets this BananaDKP to act as a server."],
					get = function() return db.mode == 1 end,
					set = function(v)
						db.mode = v and 1 or 2
					end,
				},
				host = {
					type = "text",
					name = L["Host"],
					desc = L["If acting as a client, set the host (player name) that is running the server here.\n\nNote that it's case sensitive, and it only accepts names of players who are in your guild."],
					get = function() return db.host end,
					set = function(v)
						db.host = v
					end,
					validate = function(v) return guildMemberList[v] and true or false end,
					usage = "<Player in your guild>",
					disabled = function() return db.mode ~= 2 end,
				},
			},
			order = 301,
		},
		sitouts = {
			type = "group",
			name = L["Sitouts"],
			desc = L["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."],
			disabled = function() return type(currentLeader) ~= "boolean" or not currentLeader end,
			args = {
				add = {
					type = "text",
					name = L["Add player"],
					desc = L["Adds a player to the sitout list."],
					get = false,
					usage = "<player name>",
					set = "AddSitoutPlayer",
					order = 1,
				},
				clear = {
					type = "execute",
					name = L["Clear list"],
					desc = L["Clear the sitout list."],
					func = "ClearSitoutList",
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 3,
				},
			},
			order = 302,
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return false
			end,
			hidden = notHost,
		},
		ranks = {
			type = "text",
			name = L["Ranks"],
			desc = L["Unchecked guild ranks will not show in the tooltip. People with these ranks will continue to earn and spend DKP as per usual, they will just be hidden from the display."],
			get = function(key)
				return not db.disallowRanks[key]
			end,
			set = function(key, val)
				db.disallowRanks[key] = not val
				if val then
					print(L["Members of rank %q will show in the tooltip."]:format(key))
				else
					print(L["Members of rank %q will NOT show in the tooltip."]:format(key))
				end
			end,
			order = 400,
			multiToggle = true,
			validate = guildRanks,
			disabled = function() return not IsInGuild() end,
		},
		classes = {
			type = "text",
			name = L["Classes"],
			desc = L["Unchecked classes will not be shown in the tooltip. Players of these classes will continue to earn and spend DKP as per usual, they will just be hidden from the display."],
			get = function(key)
				return not db.disallowClasses[key]
			end,
			set = function(key, val)
				db.disallowClasses[key] = not val
				if val then
					print(L["Members of class %q will show in the tooltip."]:format(key))
				else
					print(L["Members of class %q will NOT show in the tooltip."]:format(key))
				end
			end,
			order = 401,
			multiToggle = true,
			validate = guildClasses,
			disabled = function() return not IsInGuild() end,
		},
		minimap = {
			type = "toggle",
			name = L["Minimap icon"],
			desc = L["Toggle the minimap icon."],
			get = function() return not db.minimap.hide end,
			set = function(v)
				local hide = not v
				db.minimap.hide = hide
				if hide then
					icon:Hide("BananaDKP")
					print(L["BananaDKP minimap icon hidden, you can toggle it back with /bananadkp."])
				else
					icon:Show("BananaDKP")
				end
			end,
			order = 402,
			disabled = function() return not icon end,
		},
	},
}

local dew = AceLibrary("Dewdrop-2.0")
local function menu() dew:FeedAceOptionsTable(options) end
local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("BananaDKP", {
	type = "data source",
	text = "BDKP",
	icon = "Interface\\Icons\\Ability_Hunter_AspectOfTheMonkey",
	OnClick = function(self, button)
		if button == "RightButton" then
			dew:Open(self, "children", menu)
		end
	end
})

--------------------------------------------------------------------------------
-- Initialization
--

function BananaDKP:OnInitialize()
	if not _G.BananaDKPDB then _G.BananaDKPDB = {} end
	db = _G.BananaDKPDB
	for k, v in pairs({
		sitouts = {}, -- Remember this across sessions.
		mode = 2, -- 1 = server, 2 = client
		host = nil, -- No host set by default.
		disallowRanks = {},
		disallowClasses = {},
		dkp = {},
		hours = {},
		totalRaidHours = 0,
		attendance = {},
		attendanceAverageOver = 45,
		itemsLooted = {},
		minimap = {},
		lastDKPUpdate = nil,
	}) do
		if not db[k] then db[k] = v end
	end

	-- Update from r133 to r136
	-- XXX remove after next release
	for k, v in pairs(db.attendance) do
		if type(v) == "table" then
			local att = ""
			for i, bool in ipairs(v) do
				att = att .. (bool and "1" or "0")
			end
			db.attendance[k] = att
		end
	end

	-- In case the attendanceAverageOver value has changed,
	-- update attendance accordingly.
	local aAO = db.attendanceAverageOver
	for k, v in pairs(db.attendance) do
		-- Attendance tracked for longer than aAO, so trim it
		if #v > aAO then
			-- Shave from the old, since the recent data is probably more
			-- accurate
			db.attendance[k] = v:sub(1, aAO)
		
		-- Attendance not tracked for as long as aAO, so append
		elseif #v < aAO then
			local d = aAO - #v
			-- Add the new hours at the end, in case it's a slacker,
			-- he'll drop faster back down on attendance.
			db.attendance[k] = v .. ("1"):rep(d)
		end
	end

	_G["SlashCmdList"]["BDKP_SHORTHAND"] = function()
		if db.minimap.hide then
			icon:Show("BananaDKP")
			db.minimap.hide = nil
		else
			icon:Hide("BananaDKP")
			db.minimap.hide = true
		end
	end
	_G["SLASH_BDKP_SHORTHAND1"] = "/bananadkp"

	icon:Register("BananaDKP", LDB, db.minimap)

	if IsInGuild() then GuildRoster() end
end

function BananaDKP:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end
	if type(popup["BananaDKPItemQuery"]) ~= "table" then
		popup["BananaDKPItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnCancel = ignoreItem,
			OnAccept = function(self, data)
				BananaDKP:AddItem(unpack(data))
			end,
		}
	end

	for event, filter in pairs({
		CHAT_MSG_WHISPER = chatFilter,
		CHAT_MSG_WHISPER_INFORM = chatFilter,
		CHAT_MSG_RAID = chatFilter,
		CHAT_MSG_RAID_LEADER = chatFilter,
	}) do
		ChatFrame_AddMessageEventFilter(event, filter)
	end

	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")

	self:RegisterComm("BDKP-REQ",   "OnRequestComm")
	self:RegisterComm("BDKP-DKP-M", "OnMultiDKPComm")
	self:RegisterComm("BDKP-ATT-M", "OnMultiAttendanceComm")
	self:RegisterComm("BDKP-DKP-S", "OnSingleDKPComm")
	self:RegisterComm("BDKP-ATT-S", "OnSingleAttendanceComm")

	self:RebuildSitoutList()

	if db.mode == 1 then
		broadcastMulti()
		handle_Broadcast = self:ScheduleRepeatingTimer(broadcastMulti, oneHour)
	elseif db.host and (not db.lastDKPUpdate or (time() > db.lastDKPUpdate + oneHour)) then
		self:SendCommMessage("BDKP-REQ", db.host, "GUILD")
	end
end

--------------------------------------------------------------------------------
-- Comm handling
--

function BananaDKP:OnRequestComm(prefix, message, distribution, sender)
	if db.mode ~= 1 or message ~= playerName then return end
	debug("OnRequestComm(" .. sender .. ")")
	broadcastMulti()
end

function BananaDKP:OnMultiDKPComm(prefix, message, distribution, sender)
	debug("OnMultiDKPComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, dkp = self:Deserialize(message)
	if not ret then error(dkp) end
	for k, v in pairs(dkp) do
		db.dkp[k] = v
	end
	wipe(dkp)
	db.lastDKPUpdate = time()
end

function BananaDKP:OnSingleDKPComm(prefix, message, distribution, sender)
	debug("OnSingleDKPComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, tbl = self:Deserialize(message)
	if not ret then error(tbl) end
	db.dkp[tbl[1]] = tbl[2]
	db.lastDKPUpdate = time()
end

function BananaDKP:OnMultiAttendanceComm(prefix, message, distribution, sender)
	debug("OnMultiAttendanceComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, attendance = self:Deserialize(message)
	if not ret then error(attendance) end
	for k, v in pairs(attendance) do
		if v == "-" then -- All 0's
			db.attendance[k] = ("0"):rep(db.attendanceAverageOver)
		elseif v == "!" then -- All 1's
			db.attendance[k] = ("1"):rep(db.attendanceAverageOver)
		else
			db.attendance[k] = v
		end
		cache_AttendancePercent[k] = nil
	end
	wipe(attendance)
	db.lastDKPUpdate = time()
end

function BananaDKP:OnSingleAttendanceComm(prefix, message, distribution, sender)
	debug("OnSingleAttendanceComm(" .. sender .. ")")
	if db.mode == 1 or sender ~= db.host then return end
	local ret, tbl = self:Deserialize(message)
	if not ret then error(tbl) end
	db.attendance[tbl[1]] = tbl[2]
	cache_AttendancePercent[tbl[1]] = nil
	db.lastDKPUpdate = time()
end

--------------------------------------------------------------------------------
-- Ticker handling
--

local function storeTickerCounter(timeLeft)
	db.storedCounter = timeLeft - 60
	handle_MinuteTicker = BananaDKP:ScheduleTimer(storeTickerCounter, 60, db.storedCounter)
end

function BananaDKP:StartCounter()
	if handle_HourTicker then return end
	local nextUpdate = db.storedCounter or (math.random(10, 30) * 60)
	handle_HourTicker = self:ScheduleTimer("IncrementHourCounter", nextUpdate)

	if db.storedCounter then
		print(L["Starting hour counter from last session. Time until next increment is %d minutes."]:format(nextUpdate / 60))
	else
		print(L["Automatic hour counting has started, the first increment will be in %d minutes."]:format(nextUpdate / 60))
		self:Chat(L["BDKP hour counter started, whisper me 'sitout' if you are not in the raid but available on standby."], "GUILD")
		for player in pairs(db.sitouts) do
			self:Chat(L["BDKP: You are on sitout."], "WHISPER", nil, player)
		end
		db.storedCounter = nextUpdate
	end

	self:CancelTimer(handle_Broadcast, true)
	handle_Broadcast = nil
	handle_MinuteTicker = self:ScheduleTimer(storeTickerCounter, 60, nextUpdate)
	self:RegisterEvent("CHAT_MSG_LOOT")
end

function BananaDKP:StopCounter()
	if not handle_HourTicker then return end
	self:CancelTimer(handle_HourTicker)
	self:CancelTimer(handle_MinuteTicker)
	handle_HourTicker = nil
	handle_MinuteTicker = nil
	db.storedCounter = nil
	handle_Broadcast = self:ScheduleRepeatingTimer(broadcastMulti, oneHour)
	print(L["Automatic hour counting stopped."])
	
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

--------------------------------------------------------------------------------
-- Item handling
--

function BananaDKP:AddItem(player, itemLink)
	if db.mode ~= 1 then return end
	lootDialogOpen = nil
	-- DKP -25%
	local dkp = db.dkp[player] or 0
	local price = 0
	if dkp > 0 then
		price = dkp * 25 / 100
	end
	local newDkp = dkp - price
	db.dkp[player] = newDkp
	local l = db.itemsLooted[player] or 0
	db.itemsLooted[player] = l + 1
	broadcastSingle(player)
	print(L["%s looted %s, %d has been deducted from his DKP."]:format(player, itemLink, price or 0))
	popLootQueue()
end

do
	local lootSelf = (LOOT_ITEM_SELF):gsub("%%s", "(.+)")
	local lootSelfMultiple = (LOOT_ITEM_SELF_MULTIPLE):gsub("%%sx%%d", "(.+)x(%%d+)")
	local loot = (LOOT_ITEM):gsub("%%s", "(.+)")
	local lootMultiple = (LOOT_ITEM_MULTIPLE):gsub("%%s receives loot: %%sx%%d", "(.+) receives loot: (.+)x(%%d+)")
	function BananaDKP:CHAT_MSG_LOOT(_, msg)
		local player, item, quantity = playerName, nil, 1
		if msg:match(lootSelf) then
			item = msg:match(lootSelf)
		elseif msg:match(loot) then
			player, item = msg:match(loot)
		elseif msg:match(lootSelfMultiple) then
			item, quantity = msg:match(lootSelfMultiple)
		elseif msg:match(lootMultiple) then
			player, item, quantity = msg:match(lootMultiple)
		end
		if type(item) == "string" and type(player) == "string" then
			local itemName, itemLink, itemRarity = GetItemInfo(item)
			if not itemRarity or itemRarity < 4 then return end

			-- Find the item ID to make sure it's not a ignored item
			local itemId = select(3, itemLink:find("item:(%d+):"))
			if not itemId then return end
			itemId = tonumber(itemId:trim())
			if type(itemId) ~= "number" or ignoredItems[itemId] then return end

			if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
				itemLooted(player, itemLink)
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Hour handling
--

function BananaDKP:IncrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if db.dkp[input] then
			db.dkp[input] = db.dkp[input] + 1
		else
			db.dkp[input] = 1
		end
		db.hours[input] = db.hours[input] and db.hours[input] + 1 or 1
		broadcastSingle(input)
		print(L["The hour counter has been incremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			print(L["Can't increment the hour counter for %q."]:format(input))
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = 1 * getDKPMultiplier(n)
		end
		for k, v in pairs(db.sitouts) do
			if not dkpTable[k] then dkpTable[k] = 1 end
		end
		for k, v in pairs(dkpTable) do
			if not db.dkp[k] then
				db.dkp[k] = v
			else
				db.dkp[k] = db.dkp[k] + v
			end
			db.hours[k] = db.hours[k] and db.hours[k] + 1 or 1
		end
		for k in pairs(db.dkp) do
			-- We assume people have full attendance from day 1.
			if not db.attendance[k] then db.attendance[k] = ("1"):rep(db.attendanceAverageOver) end
			cache_AttendancePercent[k] = nil
			db.attendance[k] = (dkpTable[k] and "1" or "0") .. db.attendance[k]
			if #db.attendance[k] > db.attendanceAverageOver then
				db.attendance[k] = db.attendance[k]:sub(1, #db.attendance[k] - 1)
			end
		end
		db.totalRaidHours = db.totalRaidHours + 1
		wipe(dkpTable)
		broadcastMulti()
		print(L["DKP has been awarded for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			handle_HourTicker = self:ScheduleTimer("IncrementHourCounter", oneHour)
			self:CancelTimer(handle_MinuteTicker, true)
			handle_MinuteTicker = self:ScheduleTimer(storeTickerCounter, 60, oneHour)
		end
	end
end

function BananaDKP:DecrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if db.dkp[input] then
			db.dkp[input] = db.dkp[input] - 1
		end
		broadcastSingle(input)
		print(L["The hour counter has been decremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			print(L["Can't decrement the hour counter for %q."]:format(input))
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = true
		end
		for k, v in pairs(db.sitouts) do
			dbkTable[k] = true
		end
		for k, v in pairs(dkpTable) do
			if db.dkp[k] then
				db.dkp[k] = db.dkp[k] - 1
			else
				db.dkp[k] = -1
			end
		end
		wipe(dkpTable)
		broadcastMulti()
		print(L["DKP has been decremented by 1 for all players in the raid and on the sitout list."])
	end
end

--------------------------------------------------------------------------------
-- Group and guild handling
--

do
	local grouped = nil
	local lastNumRaid = nil
	local lastNumGroup = nil
	function BananaDKP:RAID_ROSTER_UPDATE()
		local n = GetNumRaidMembers()
		if not lastNumRaid or lastNumRaid ~= n then updateDisplayedNames = true end
		lastNumRaid = n
		if not grouped and n > 0 then
			grouped = true
		elseif grouped and n == 0 then
			self:StopCounter()
			self:ClearSitoutList()
			grouped = nil
		end
		
		if n > 0 and not handle_HourTicker and db.storedCounter then
			self:StartCounter()
		end
	end
	function BananaDKP:PARTY_MEMBERS_CHANGED()
		local n = GetNumPartyMembers()
		if not lastNumGroup or lastNumGroup ~= n then updateDisplayedNames = true end
		lastNumGroup = n
		if not grouped and n > 0 then
			grouped = true
		elseif grouped and n == 0 and not UnitInRaid("player") then
			grouped = nil
			self:StopCounter()
			self:ClearSitoutList()
		end
	end
end

do
	-- This is just run on login, since I can't seem to get a reliable hit from
	-- RAID_ROSTER_UPDATE, PLAYER_LOGIN or PLAYER_ENTERING_WORLD on
	-- GetNumRaidMembers ...
	local f = CreateFrame("Frame")
	local total = 0
	local seconds = 0
	f:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total > 1 then
			seconds = seconds + 1
			if seconds == 10 or GetNumRaidMembers() > 0 then
				BananaDKP:RAID_ROSTER_UPDATE()
				f:SetScript("OnUpdate", nil)
			end
			total = 0
		end
	end)
end

do
	local updating = nil
	function BananaDKP:GUILD_ROSTER_UPDATE()
		if updating then return end
		updating = true
		self:UnregisterEvent("GUILD_ROSTER_UPDATE")
		wipe(guildRanks)
		for i = 1, GuildControlGetNumRanks() do
			table.insert(guildRanks, GuildControlGetRankName(i))
		end
		local offline = GetGuildRosterShowOffline()
		local selection = GetGuildRosterSelection()
		SetGuildRosterShowOffline(true)
		local classes = {}
		for i = 1, GetNumGuildMembers() do
			local name, rank, _, _, localizedClass, _, _, _, _, _, class = GetGuildRosterInfo(i)
			if name then
				guildMemberList[name] = class
				guildMemberRanks[name] = rank
				guildMemberClasses[name] = localizedClass
				classes[localizedClass] = true
			end
		end
		SetGuildRosterShowOffline(offline)
		SetGuildRosterSelection(selection)

		wipe(guildClasses)
		for k in pairs(classes) do
			table.insert(guildClasses, k)
		end
		wipe(classes)

		updateDisplayedNames = true
		updating = nil
		self:RegisterEvent("GUILD_ROSTER_UPDATE")
	end
end

--------------------------------------------------------------------------------
-- Chat filters
--

local function tablePrinter(t, recipient)
	for i, v in ipairs(t) do
		if type(v) == "string" then
			BananaDKP:Chat(L["BDKP: %s."]:format(v), "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

local function bidTimeout()
	wipe(tmpBidTable)
	for k in pairs(bidders) do table.insert(tmpBidTable, k) end
	if #tmpBidTable > 0 then
		table.sort(tmpBidTable, dkpSorter)
		local winner = tmpBidTable[1]
		if #tmpBidTable > 1 and bidders[winner] == bidders[tmpBidTable[2]] then
			local more = true
			while more do
				table.remove(tmpBidTable, #tmpBidTable)
				more = bidders[tmpBidTable[#tmpBidTable]] ~= bidders[winner]
			end
			BananaDKP:Chat(L["Tie: %s."]:format(table.concat(tmpBidTable, ", ")), "RAID")
		else
			BananaDKP:Chat(L["%s wins (%d dkp)!"]:format(winner, bidders[winner]), "RAID")
		end
	end
	takingBids = nil
	handle_BidTimer = nil
end

local function newBidder(bidder)
	bidders[bidder] = getModifiedDKP(bidder) or 0
	BananaDKP:CancelTimer(handle_BidTimer, true)
	handle_BidTimer = BananaDKP:ScheduleTimer(bidTimeout, 15)
end

chatHandlers.CHAT_MSG_WHISPER = function(msg, author)
	if db.mode == 2 then return end
	if msg:find("^bdkp") then
		local command = select(3, msg:find("^bdkp%s?(.*)"))
		local input = command and command:len() > 0 and command:gsub("^%l", string.upper) or author
		local dkp = getDKP(input)
		if type(dkp) == "number" then
			local realDkp = getModifiedDKP(input)
			local attendance = getAttendancePercent(input)
			local dkpPercent = getDKPPercent(input)
			local format = L["BDKP: DKP for %s: %s (%s total). 45-hour attendance is at %d%%, which means DKP is worth %d%% of its total value for purposes of bidding."]
			local ret = format:format(input, tostring(realDkp), tostring(dkp), attendance, dkpPercent)
			BananaDKP:Chat(ret, "WHISPER", nil, author)
		else
			BananaDKP:Chat(L["BDKP: Unknown command or player not found."], "WHISPER", nil, author)
		end
		return true
	elseif takingBids and msg:find("^bid") then
		newBidder(author)
		return true
	elseif msg == "sitout" then
		BananaDKP:ProcessSitoutWhisper(author)
		return true
	end
end

chatHandlers.CHAT_MSG_WHISPER_INFORM = function(msg, recipient)
	if msg:find("^BDKP") then return true end
end

chatHandlers.CHAT_MSG_RAID = function(msg, author)
	if db.mode == 2 then return end
	if author == playerName and msg:find(L["^bid:"]) then
		takingBids = true
		wipe(bidders)
	elseif takingBids and msg:find(L["^bid"]) then
		newBidder(author)
	end
end
chatHandlers.CHAT_MSG_RAID_LEADER = chatHandlers.CHAT_MSG_RAID

function BananaDKP:Chat(text, chatType, language, destination)
	if ChatThrottleLib then
		ChatThrottleLib:SendChatMessage("BULK", "BananaDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

--------------------------------------------------------------------------------
-- LDB Plugin Tooltip
--

do
	local names = {}
	local nameFormat = "%s |cff555555(%d%%)|r"

	local realm = GetRealmName()
	local function shouldShow(name)
		if guildMemberList[name] and
			db.disallowRanks[guildMemberRanks[name]] or
			db.disallowClasses[guildMemberClasses[name]] then
			return nil
		end
		-- Don't show players from other realms.
		local n, r = UnitName(name)
		-- Apparently the second return value can be "" for people from our own realm, not just nil, like wowwiki suggests.
		if n and r and r:trim():len() > 0 and r ~= realm then
			return nil
		end
		return true
	end

	function LDB.OnTooltipShow(tt)
		tt:AddLine("BananaDKP")
		if db.mode == 2 and db.lastDKPUpdate then
			tt:AddDoubleLine(L["Updated"], date("%d. %b@%H:%M", db.lastDKPUpdate), 1,1,0, 1,1,1)
		end

		if updateDisplayedNames then
			wipe(names)
			if GetNumRaidMembers() > 0 then
				for i = 1, GetNumRaidMembers() do
					local n = GetRaidRosterInfo(i)
					if n and shouldShow(n) then
						table.insert(names, n)
					end
				end
			elseif GetNumPartyMembers() > 0 then
				table.insert(names, playerName)
				for i = 1, GetNumPartyMembers() do
					local n = UnitName("party" .. i)
					if n and shouldShow(n) then
						table.insert(names, n)
					end
				end
			else
				for k in pairs(guildMemberList) do
					if shouldShow(k) then
						local d = getDKP(k)
						if d then table.insert(names, k) end
					end
				end
			end
			updateDisplayedNames = nil
			table.sort(names, dkpSorter)
		end

		if #names > 0 then
			tt:AddLine(" ")
			tt:AddDoubleLine(L["Player"], L["DKP"], 1,1,1, 1,1,1)
			for i, v in ipairs(names) do
				local totalDkp = getDKP(v)
				if totalDkp and totalDkp > 0 then
					local realDkp = getModifiedDKP(v)
					local attendance = getAttendancePercent(v)
					if v == playerName then
						tt:AddDoubleLine(nameFormat:format(v, attendance), realDkp, 1,0,0, 1,0,0)
					else
						tt:AddDoubleLine(nameFormat:format(coloredName[v], attendance), realDkp, 1,1,1, 1,1,1)
					end
				end
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Sitout list handling
--

function BananaDKP:ProcessSitoutWhisper(player)
	if not guildMemberList[player] then
		print(L["%q tried to add himself to our sitout list, but he's not in our guild."]:format(player))
		return
	end
	if db.sitouts[player] then
		self:RemoveSitoutPlayer(player)
	else
		self:AddSitoutPlayer(player)
	end
end

function BananaDKP:ClearSitoutList()
	wipe(db.sitouts)
	self:RebuildSitoutList()
end

function BananaDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	print(L["%q added to the sitout list."]:format(name))
	self:Chat(L["BDKP: You have been ADDED to the sitout list. Whisper me again to remove yourself."], "WHISPER", nil, name)
	self:RebuildSitoutList()
end

function BananaDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	print(L["%q removed from the sitout list."]:format(name))
	self:Chat(L["BDKP: You have been REMOVED from the sitout list. Whisper me again to add yourself."], "WHISPER", nil, name)
	self:RebuildSitoutList()
end

function BananaDKP:RebuildSitoutList()
	local opts = options.args.sitouts.args

	for k, v in pairs(opts) do
		if k ~= "add" and k ~= "clear" and k ~= "spacer" then
			opts[k].hidden = true
		end
	end

	for k, v in pairs(db.sitouts) do
		if not opts[k] then
			opts[k] = {
				type = "execute",
				name = coloredName[k],
				desc = L["Remove %s."]:format(k),
				func = "RemoveSitoutPlayer",
				passValue = k,
				order = 100,
			}
		end
		opts[k].hidden = nil
	end
end

